Module DS1307

// choose hardware or software I2C (default is hardware I2C)
#option DS1307_SI2C = false
#if IsOption (DS1307_SI2C) And Not (DS1307_SI2C in (true, false))
   #error DS1307_SI2C, "Invalid Option, DS1307_SI2C must be true or false"
#endIf

// import modules...
Include "Convert.bas" // BCD routines
Include "RTC.bas"     // time and date structures

// software or hardware I2C... 
#if (_mssp > 0) And Not DS1307_SI2C
   Include "I2C.bas"
#else
   Include "SI2C.bas"
#endif

// the module uses 'generic' I2C sub and function calls which are mapped onto either 
// hardware or software I2C - depending on the #option DS1307_SI2C...
Dim 
   I2CWriteByte As WriteByte,
   I2CReadByte As ReadByte,
   I2CInitialize As Initialize,
   I2CStart As Start,
   I2CStop As Stop,
   I2CRestart As Restart

// local module constants...
Const
   RTC_DEV_ADDR    = $D0,   // DS1307 device Address
   RTC_TIME        = $00,   // Address for time
   RTC_WDAY        = $03,   // Address for day of week number
   RTC_DATE        = $04,   // Address for date
   RTC_CONTROL     = $07,   // Address for control byte
   RTC_NVRAM       = $08,   // NVRAM start
   RTC_NVRAM_MAX   = $38    // NVRAM address max (56 decimal)

// public constants - used in calls to ReadControl() and WriteControl()...
Public Const   
   RTC_SQWE        = $10,   // Squarewave enable
   RTC_SQWF_1HZ    = $0,    // Squarewave frequency  1Hz
   RTC_SQWF_4KHz   = $1,    // Squarewave frequency  4.096KHz
   RTC_SQWF_8KHz   = $2,    // Squarewave frequency  8.192KHz 
   RTC_SQWF_32KHz  = $3,    // Squarewave frequency 32.768KHz  
   RTC_SQWOUT_H    = $80    // Set Squarewave output to high when disabled

// public variables...
Public Dim
   Time As TTime,    
   Date As TDate
{
****************************************************************************
* Name    : RTCStartWrite (PRIVATE)                                        *
* Purpose : Starts a RTC write                                             *
****************************************************************************
}
Sub RTCStartWrite(Address As Byte) 
   I2CStart
   I2CWriteByte(RTC_DEV_ADDR)
   I2CWriteByte(Address)        
End Sub
{
****************************************************************************
* Name    : RTCStartRead (PRIVATE)                                         *
* Purpose : Starts a RTC read                                              *
****************************************************************************
}
Sub RTCStartRead(Address As Byte)
   RTCStartWrite(Address)
   I2CRestart
   I2CWriteByte(RTC_DEV_ADDR + 1)
End Sub
{
****************************************************************************
* Name    : RTCStop (PRIVATE)                                              *
* Purpose : Stops RTC read or write                                        *
****************************************************************************
}
Sub RTCStop()
   Acknowledge(I2C_NOT_ACKNOWLEDGE)
   Stop
End Sub
{
****************************************************************************
* Name    : ReadRegister (PRIVATE)                                         *
* Purpose : Read DS1307 register ($00..$3F)                                *
****************************************************************************
}         
Function ReadRegister(pAddress As Byte) As Byte
   RTCStartRead(pAddress)
   Result = I2CReadByte(I2C_NOT_ACKNOWLEDGE)
   RTCStop
End Function
{
****************************************************************************
* Name    : WriteRegister (PRIVATE)                                        *
* Purpose : Write to DS1307 register ($00..$3F)                            *
****************************************************************************
}         
Sub WriteRegister(pAddress As Byte, pData As Byte)
   RTCStartWrite(pAddress)
   I2CWriteByte(pData)
   RTCStop
End Sub
{
****************************************************************************
* Name    : ReadControl                                                    *
* Purpose : Returns the RTC control byte ($07)                             *
*         : See the control constants RTC_SQWXXXX above                    *
****************************************************************************
}         
Public Function ReadControl() As  Byte
   Result = ReadRegister(RTC_CONTROL)
End Function  
{
****************************************************************************
* Name    : WriteControl                                                   *
* Purpose : Writes the RTC Control byte ($07)                              *
*         : See the control constants RTC_SQWXXXX above                    *
****************************************************************************
}         
Public Sub WriteControl(pData As Byte)
   WriteRegister(RTC_CONTROL, pData)
End Sub
{
****************************************************************************
* Name    : WriteByte                                                      *
* Purpose : Writes a byte to NV Ram. The RAM is physically located at      *
*         : $08..$3F - this routine uses address $00..$37 (56 bytes)       *
****************************************************************************
}         
Public Sub WriteByte(pAddress As Byte, pData As Byte)
   If pAddress < RTC_NVRAM_MAX Then
      WriteRegister(pAddress + RTC_NVRAM, pData)
   EndIf
End Sub
{
****************************************************************************
* Name    : ReadByte                                                       *
* Purpose : Reads a byte from NV Ram. The RAM is physically located at     *
*         : $08..$3F - this routine uses address $00..$37 (56 bytes)       *
****************************************************************************
}         
Public Function ReadByte(pAddress As Byte) As Byte
   Result = ReadRegister(pAddress + RTC_NVRAM)
End Function 
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Returns Time in the referenced variable                        *
****************************************************************************
}         
Sub ReadItem(ByRef pTime As TTime)
   RTCStartRead(RTC_TIME)
   Time.Second  = BCDToDec(I2CReadByte(I2C_ACKNOWLEDGE))
   Time.Minute  = BCDToDec(I2CReadByte(I2C_ACKNOWLEDGE))
   Time.Hour  = I2CReadByte(I2C_NOT_ACKNOWLEDGE)
   RTCStop
   Time.Mode.Is12Hour = Time.Hour.Booleans(6)
   If Time.Mode.Is12Hour Then
      Time.Mode.PM = Time.Hour.Booleans(5)
      Time.Hour = Time.Hour And $1F
   EndIf   
   Time.Hour = BCDToDec(Time.Hour)
   pTime = Time
End Sub
{
****************************************************************************
* Name    : ReadItem (OVERLOAD)                                            *
* Purpose : Returns Date in the referenced variable                        *
****************************************************************************
}
Sub ReadItem(ByRef pDate As TDate)
   RTCStartRead(RTC_DATE)
   Date.Day   = BCDToDec(I2CReadByte(I2C_ACKNOWLEDGE))
   Date.Month = BCDToDec(I2CReadByte(I2C_ACKNOWLEDGE))
   Date.Year  = BCDToDec(I2CReadByte(I2C_NOT_ACKNOWLEDGE))
   RTCStop

   // day of week...
   RTCStartRead(RTC_WDAY)
   Date.DayOfWeek = BCDToDec(I2CReadByte(I2C_NOT_ACKNOWLEDGE))
   RTCStop 
   pDate = Date  
End Sub
{
****************************************************************************
* Name    : Read                                                           *
* Purpose : Returns Time or Date in the referenced variable                *
****************************************************************************
}         
Public Compound Sub Read(ReadItem)
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Sets the Time in BCD format                                    *
****************************************************************************
}         
Sub WriteItem(pTime As TTime)
   Time = pTime
   pTime.Hour = DecToBCD(pTime.Hour) 
   If pTime.Mode.Is12Hour Then
      pTime.Hour.Booleans(5) = pTime.Mode.PM
      pTime.Hour.Booleans(6) = pTime.Mode.Is12Hour
   EndIf
   RTCStartWrite(RTC_TIME)
   I2CWriteByte(DecToBCD(pTime.Second))
   I2CWriteByte(DecToBCD(pTime.Minute))
   I2CWriteByte(pTime.Hour)
   RTCStop
End Sub   
{
****************************************************************************
* Name    : WriteItem (OVERLOAD)                                           *
* Purpose : Sets the Date in BCD format                                    *
****************************************************************************
}         
Sub WriteItem(pDate As TDate)
   Date = pDate
   RTCStartWrite(RTC_DATE)
   I2CWriteByte(DecToBCD(pDate.Day))
   I2CWriteByte(DecToBCD(pDate.Month))
   I2CWriteByte(DecToBCD(pDate.Year))
   RTCStop

   // day of week...
   RTCStartWrite(RTC_WDAY)
   I2CWriteByte(DecToBCD(pDate.DayOfWeek))
   RTCStop   
End Sub   
{
****************************************************************************
* Name    : Write                                                          *
* Purpose : Sets the Time or Date                                          *
****************************************************************************
}         
Public Compound Sub Write(WriteItem)
{
****************************************************************************
* Name    : Enabled                                                        *
* Purpose : Returns true if the RTC has been set up and is running         *
****************************************************************************
}         
Public Function Enabled() As Boolean
   Result = (ReadRegister(RTC_TIME) And $80) = 0  
End Function
{
****************************************************************************
* Name    : Initialize (PRIVATE)                                           *
* Purpose : Initializes the module                                         *
****************************************************************************
}         
Sub Initialize()
   Date.Day = 1        // 1st
   Date.Month = 1      // January
   Date.Year = 0       // 2000
   Date.DayOfWeek = 6  // Saturday
   Time.Hour = 0       // 12 midnight
   Time.Minute = 0
   Time.Second = 0
   Clear(Time.Mode)    // 24 hour clock
   I2CInitialize
End Sub

// initialize the module...
Initialize
