{
*************************************************************************
*  Name    : MCP23S17_MULT.BAS                                          *
*  Author  : Ahmed Lazreg  (octal@pocketmt.com)                         *
*            http://www.pocketmt.com                                    *
*  Notice  : Copyright (c) 2006 Pocket MicroTechnics                    *
*          : All Rights Reserved                                        *
*  Date    : 01/03/2008                                                 *
*  Version : 1.0                                                        *
*  Modified:                                                            *
*  Notes   : This module handles multiple MCP23S17 - 16 pins I/O port   *
*            expander from Microchip as 2x8 bits PortA/PortB.           *
*            The module shares the same Clk/SDI/SDO/CS/RST pins,        *
*            selection is done using chip address pins.                 *    
*************************************************************************
}


Module MCP23S17_MULT

// hardware or software SPI...
#if IsOption(MCPSPI_SOFTWARE) And Not (MCPSPI_SOFTWARE in (true, false))
   #error MCPSPI_SOFTWARE, "Invalid option, must be TRUE or FALSE."
#endif
#option MCPSPI_SOFTWARE = false

// enable delays...
#if IsOption(MCP23S17_DELAYS) And Not (MCP23S17_DELAYS in (true, false))
   #error MCP23S17_DELAYS, "Invalid option, must be TRUE or FALSE."
#endif
#option MCP23S17_DELAYS = false

#option MCP23S17_DELAY_US = 10
Const MCP23S17_DELAY_US_ = MCP23S17_DELAY_US

#if MCPSPI_SOFTWARE=true // use software SPI...
Include "sspi.bas"
#else // use hardware SPI...
Include "SPI.BAS"
#EndIf

#Option MCP_CS  = PORTC.1
//#Option MCP_RST = PORTC.2   // may be commented If RST is Not used

// validate SCK pin...
#If IsOption(MCP_CS) And Not IsValidPortPin(MCP_CS) 
   #Error MCP_CS, "Invalid option. MCP_CS must be a valid port pin."
#EndIf

// validate RST pin...
#If IsOption(MCP_RST) And Not IsValidPortPin(MCP_RST) 
   #Error MCP_RST, "Invalid option. MCP_RST must be a valid port pin."
#EndIf

Public Dim 
   CS As MCP_CS.MCP_CS@   // Chip select

#if IsOption(MCP_RST)
Public Dim
   RST As MCP_RST.MCP_RST@    // RST pin
#endif

//----------------------------------------------------------------------------
//Please ignore these options ... only here for the GLCD through SPI stuff...
#option MCPPortA0 = 0
#option MCPPortA1 = 1
#option MCPPortA2 = 2
#option MCPPortA3 = 3
#option MCPPortA4 = 4
#option MCPPortA5 = 5
#option MCPPortA6 = 6
#option MCPPortA7 = 7

#option MCPPortB0 = 8
#option MCPPortB1 = 9
#option MCPPortB2 = 10
#option MCPPortB3 = 11
#option MCPPortB4 = 12
#option MCPPortB5 = 13
#option MCPPortB6 = 14
#option MCPPortB7 = 15
//----------------------------------------------------------------------------


Public Const
  IODIRA   = $00, 
  IODIRB   = $01,
  IPOLA    = $02,
  IPOLB    = $03,
  GPINTENA = $04,
  GPINTENB = $05,
  DEFVALA  = $06,
  DEFVALB  = $07,
  INTCONA  = $08,
  INTCONB  = $09,
  IOCONA   = $0A,
  IOCONB   = $0B,
  GPPUA    = $0C,
  GPPUB    = $0D,
  INTFA    = $0E,
  INTFB    = $0F,
  INTCAPA  = $10,
  INTCAPB  = $11,
  GPIOA    = $12,
  GPIOB    = $13,
  OLATA    = $14,
  OLATB    = $15 

//Control Byte For SPI operations [0100][A2][A1][A0][R=1/W=0]
Private Const CONTROL_BYTE_MASK_R As Byte = %01000001
Private Const CONTROL_BYTE_MASK_W As Byte = %01000000

Private Const CONTROL_IOCON = %00011000 // Init value for IOCON - bits details in following comment

{
 Reg IOCON = [BANK][MIRROR][SEQOP][DISSLW][HAEN][ODR][INTPOL][-Unused]

 bit 7 BANK: Controls how the registers are addressed
             1 = The registers associated with each port are separated into different banks
             0 = The registers are in the same bank (addresses are sequential)
 bit 6 MIRROR: INT Pins Mirror bit
             1 = The INT pins are internally connected
             0 = The INT pins are not connected. INTA is associated with PortA and INTB is associated with PortB
 bit 5 SEQOP: Sequential Operation mode bit.
             1 = Sequential operation disabled, address pointer does not increment.
             0 = Sequential operation enabled, address pointer increments.
 bit 4 DISSLW: Slew Rate control bit for SDA output.
             1 = Slew rate disabled.
             0 = Slew rate enabled.
 bit 3 HAEN: Hardware Address Enable bit (MCP23S17 only).
             Address pins are always enabled on MCP23017.
             1 = Enables the MCP23S17 address pins.
             0 = Disables the MCP23S17 address pins.
 bit 2 ODR: This bit configures the INT pin as an open-drain output.
             1 = Open-drain output (overrides the INTPOL bit).
             0 = Active driver output (INTPOL bit sets the polarity).
 bit 1 INTPOL: This bit sets the polarity of the INT output pin.
             1 = Active-high.
             0 = Active-low.
 bit 0 Unimplemented: Read as 0.

}

{
****************************************************************************
* Name    : WriteReg                                                       *
* Purpose : Writes pData byte to register address given by pRegAddress     *
*           pChipAddress is the address of concerned MCP23S17 chip         *
****************************************************************************
} 
Public Sub WriteReg(ByVal pChipAddress, pRegAddress As Byte, pData As Byte)
    Dim CntByte As Byte
    CntByte = CONTROL_BYTE_MASK_W Or (pChipAddress << 1) 
    CS = 0
    #if MCP23S17_DELAYS
        DelayUS(MCP23S17_DELAY_US_)  'delays can/should be added if 
                     'you have problems at High Clock values
    #endif
    WriteByte(CntByte)
    WriteByte(pRegAddress)
    WriteByte(pData)
    CS = 1
    #if MCP23S17_DELAYS
        DelayUS(MCP23S17_DELAY_US_)  'delays can/should be added if 
                     'you have problems at High Clock values
    #endif
End Sub
{
****************************************************************************
* Name    : ReadReg                                                        *
* Purpose : Read register address given by pRegAddress                     *
****************************************************************************
} 
Public Function ReadReg(ByVal pChipAddress As Byte, ByVal pRegAddress As Byte) As Byte
    Dim CntByte As Byte
    CntByte = CONTROL_BYTE_MASK_R Or (pChipAddress << 1) 
    CS = 0
    #if MCP23S17_DELAYS
        DelayUS(MCP23S17_DELAY_US_)  'delays can/should be added if 
                     'you have problems at High Clock values
    #endif   
    WriteByte(CntByte)
    WriteByte(pRegAddress)
    Result = ReadByte()
    CS = 1  
    #if MCP23S17_DELAYS
        DelayUS(MCP23S17_DELAY_US_)  'delays can/should be added if 
                     'you have problems at High Clock values
    #endif 
End Function
{
****************************************************************************
* Name    : ReadPortA                                                      *
* Purpose : Returns PortA reg value                                        *
****************************************************************************
} 
Public Inline Function ReadPortA(ByVal pChipAddress As Byte) As Byte
    Result = ReadReg(pChipAddress, GPIOA)
End Function
{
****************************************************************************
* Name    : ReadPortB                                                      *
* Purpose : Returns PortB reg value                                        *
****************************************************************************
} 
Public Inline Function ReadPortB(ByVal pChipAddress As Byte) As Byte
    Result = ReadReg(pChipAddress, GPIOB)
End Function
{
****************************************************************************
* Name    : ReadPinPortA                                                   *
* Purpose : Returns Value of a specified PortA Pin                         *
****************************************************************************
} 
Public Inline Function ReadPinPortA(ByVal pChipAddress As Byte, pPin As Byte) As Bit
    Dim tmp As Byte
    tmp = ReadReg(pChipAddress, GPIOA)
    Result = tmp.Bits(pPin)
End Function
{
****************************************************************************
* Name    : ReadPinPortB                                                   *
* Purpose : Returns Value of a specified PortB Pin                         *
****************************************************************************
} 
Public Inline Function ReadPinPortB(ByVal pChipAddress As Byte, pPin As Byte) As Bit
    Dim tmp As Byte
    tmp = ReadReg(pChipAddress, GPIOB)
    Result = tmp.Bits(pPin)
End Function
{
****************************************************************************
* Name    : WritePortA                                                     *
* Purpose : Writes PortA Latch register with pData value                   *
****************************************************************************
} 
Public Inline Sub WritePortA(ByVal pChipAddress, pData As Byte)
    WriteReg(pChipAddress, OLATA, pData) // I prefer to write to latch instead of direct port 
                                         // This is safer in case if some pins were inputs    
End Sub
{
****************************************************************************
* Name    : WritePortB                                                     *
* Purpose : Writes PortB Latch register with pData value                   *
****************************************************************************
} 
Public Inline Sub WritePortB(ByVal pChipAddress, pData As Byte)
    WriteReg(pChipAddress, OLATB, pData) // I prefer to write to latch instead of direct port 
                                         // This is safer in case if some pins were inputs   
End Sub
{
****************************************************************************
* Name    : SetPinPortA                                                    *
* Purpose : Sets PortA pin given by pPin to High state                     *
****************************************************************************
} 
Public Sub SetPinPortA(ByVal pChipAddress, pPin As Byte)
    Dim tmp As Byte
    tmp = ReadPortA(pChipAddress) Or (1<<pPin)
    WriteReg(pChipAddress, OLATA, tmp) // I prefer to write to latch instead of direct port 
                                         // This is safer in case if some pins were inputs    
End Sub
{
****************************************************************************
* Name    : SetPinPortB                                                    *
* Purpose : Sets PortB pin given by pPin to High state                     *
****************************************************************************
} 
Public Sub SetPinPortB(ByVal pChipAddress, pPin As Byte)
    Dim tmp As Byte
    tmp = ReadPortB(pChipAddress) Or (1<<pPin)
    WriteReg(pChipAddress, OLATB, tmp) // I prefer to write to latch instead of direct port 
                                         // This is safer in case if some pins were inputs    
End Sub
{
****************************************************************************
* Name    : SetPinPortAB                                                   *
* Purpose : Sets PortA or PortB pin given by pPin to High state            *
*           Pins are numbered from 0 to 15.                                * 
****************************************************************************
} 
Public Sub SetPinPortAB(ByVal pChipAddress, pPin As Byte)
    If pPin<8 Then
      SetPinPortA(pChipAddress, pPin)
    Else
      SetPinPortB(pChipAddress, pPin)
    EndIf
End Sub
{
****************************************************************************
* Name    : ClearPinPortA                                                    *
* Purpose : Sets PortA pin given by pPin to Low state                     *
****************************************************************************
} 
Public Sub ClearPinPortA(ByVal pChipAddress, pPin As Byte)
    Dim tmp As Byte
    tmp = ReadPortA(pChipAddress) And Not (1<<pPin)
    WriteReg(pChipAddress, OLATA, tmp) // I prefer to write to latch instead of direct port 
                                         // This is safer in case if some pins were inputs    
End Sub
{
****************************************************************************
* Name    : ClearPinPortB                                                 *
* Purpose : Sets PortB pin given by pPin to Low state                     *
****************************************************************************
} 
Public Sub ClearPinPortB(ByVal pChipAddress, pPin As Byte)
    Dim tmp As Byte
    tmp = ReadPortB(pChipAddress) And Not (1<<pPin)
    WriteReg(pChipAddress, OLATB, tmp) // I prefer to write to latch instead of direct port 
                                         // This is safer in case if some pins were inputs    
End Sub
{
****************************************************************************
* Name    : ClearPinPortAB                                                   *
* Purpose : Clears PortA or PortB pin given by pPin to low state           *
*           Pins are numbered from 0 to 15.                                * 
****************************************************************************
} 
Public Sub ClearPinPortAB(ByVal pChipAddress, pPin As Byte)
    If pPin<8 Then
      ClearPinPortA(pChipAddress, pPin)
    Else
      ClearPinPortB(pChipAddress, pPin)
    EndIf
End Sub
{
****************************************************************************
* Name    : SetTrisPortA                                                   *
* Purpose : Sets PortA pins direction 0=Output 1=Input                     *
****************************************************************************
} 
Public Inline Sub SetTrisPortA(ByVal pChipAddress, pDirection As Byte)
    WriteReg(pChipAddress, IODIRA, pDirection) 
End Sub
{
****************************************************************************
* Name    : SetTrisPortB                                                   *
* Purpose : Sets PortB pins direction 0=Output 1=Input                     *
****************************************************************************
} 
Public Inline Sub SetTrisPortB(ByVal pChipAddress, pDirection As Byte)
    WriteReg(pChipAddress, IODIRB, pDirection)
End Sub
{
****************************************************************************
* Name    : GetTrisPortA                                                   *
* Purpose : Gets PortA pins direction 0=Output 1=Input                     *
****************************************************************************
} 
Public Inline Function GetTrisPortA(ByVal pChipAddress As Byte) As Byte
    result = ReadReg(pChipAddress, IODIRA) 
End Function
{
****************************************************************************
* Name    : GetTrisPortB                                                   *
* Purpose : Gets PortB pins direction 0=Output 1=Input                     *
****************************************************************************
} 
Public Inline Function GetTrisPortB(ByVal pChipAddress As Byte) As Byte
    result = ReadReg(pChipAddress, IODIRB) 
End Function
{
****************************************************************************
* Name    : SetPinTrisPortA                                                *
* Purpose : Sets the Tris of a Single Pin from PortA pins  0=Output 1=Input*
****************************************************************************
} 
Public Sub SetPinTrisPortA(ByVal pChipAddress As Byte, pPin As Byte, pDirection As Bit)
    Dim PortTris As Byte
    PortTris = ReadReg(pChipAddress, IODIRA)
    PortTris.Bits(pPin) = pDirection
    WriteReg(pChipAddress, IODIRA, PortTris)
End Sub
{
****************************************************************************
* Name    : SetPinTrisPortB                                                *
* Purpose : Sets the Tris of a Single Pin from PortB pins  0=Output 1=Input*
****************************************************************************
} 
Public Sub SetPinTrisPortB(ByVal pChipAddress As Byte, pPin As Byte, pDirection As Bit)
    Dim PortTris As Byte
    PortTris = ReadReg(pChipAddress, IODIRB)
    PortTris.Bits(pPin) = pDirection
    WriteReg(pChipAddress, IODIRB, PortTris)
End Sub
{
****************************************************************************
* Name    : SetPinTrisPortAB                                               *
* Purpose : Sets the Tris of a Single Pin from PortA or PortB              *
*          0=Output 1=Input   Pins are numbered from 0 to 15               *
****************************************************************************
} 
Public Sub SetPinTrisPortAB(ByVal pChipAddress As Byte, pPin As Byte, pDirection As Bit)
    If pPin<8 Then
      SetPinTrisPortA(pChipAddress, pPin, pDirection)
    Else
      SetPinTrisPortB(pChipAddress, pPin, pDirection)
    End If
End Sub
{
****************************************************************************
* Name    : SetPullUpsPortA                                                *
* Purpose : Sets the Pullups resistors  for any Pin of PortA               *
*           Pullups can be enabled/disabled independently for any pin of   *
*           the port 0=Disabled  1=Enabled                                 *
****************************************************************************
} 
Public Inline Sub SetPullUpsPortA(ByVal pChipAddress, pPullUps As Byte)
    WriteReg(pChipAddress, GPPUA, pPullUps)
End Sub
{
****************************************************************************
* Name    : SetPullUpsPortB                                                *
* Purpose : Sets the Pullups resistors  for any Pin of PortB               *
*           Pullups can be enabled/disabled independently for any pin of   *
*           the port 0=Disabled  1=Enabled                                 *
****************************************************************************
} 
Public Inline Sub SetPullUpsPortB(ByVal pChipAddress, pPullUps As Byte)
    WriteReg(pChipAddress, GPPUB, pPullUps)
End Sub
{
****************************************************************************
* Name    : InitChip                                                       *
* Purpose : Initialize ALL the MCP23S17 CHIP MODULES on the bus            *
****************************************************************************
} 
Public Sub InitChip()
#if IsOption(MCP_RST)    
    // Reset the Chips
    RST = 0 
    DelayMS(10)
    RST = 1
    DelayMS(10)
#EndIf       
    // At initialization, IOCON.HAEN is not set, so chip Addresses are not
    // active. All chips are considered with CHIP_ADDRESS set to "000" 
    // regardless of the state of their Adress [A2 A1 A0] pins.
    // After initialization, once IOCON.HAEN enabled, MCPCHIP_ADDRESS is 
    // taken into account by the chip.
    WriteReg(%000, IOCONA, CONTROL_IOCON) // writing to IOCONA writes also to IOCONB     
    DelayMS(10)
End Sub
{
****************************************************************************
* Name    : Initialize                                                     *
* Purpose : Initializes module and included modules                        *
****************************************************************************
} 
Public Sub Initialize()  

#If MCPSPI_SOFTWARE    
    SSPI.Initialize
    SSPI.SetClock(spiIdleHigh)
#Else
    SPI.SetAsMaster(spiOscDiv4)
    SPI.SetClock(spiIdleHigh, spiRisingEdge)
    SPI.SetSample(spiSampleMiddle)
    EnableSPI()    
#EndIf    
    Output(CS)     
#If IsOption(MCP_RST)        
    Output(RST)
    High(RST)
#EndIf       
    High(CS)          
    InitChip()      
End Sub

Initialize()

