{
********************************************************************************
*  Name    : 8 Servo Controller Module                                         *
*  Author  : AndyO                                                             *
*  Notice  : Copyright (c) 2010 AndyO                                          *
*          : All Rights Reserved                                               *
*  Date    : 07/03/2010                                                        *
*  Version : 1.0                                                               *
*  Notes   : Based on code posted by Mike McLaren in the Swordfish forum:      *
*          : http://www.sfcompiler.co.uk/forum/viewtopic.php?t=461&start=0     *
*          :                                                                   *
*          : Controls 1 to 8 hobby servos connected to PortB:                  *
*          : Servo 0 connected to PortB.0, Servo 1 connected to PortB.1, etc.  *
*          :                                                                   *
*          : Servos positioned by sending pulses between 600uS and 2400uS long *
*          : every 20mS.  Pulse length of 1500uS = approximate halfway point.  *
*          :                                                                   *
*          : Set servo positions by loading desired pulse length (in uS) into  *
*          : "ServoPosition" array e.g. "ServoPosition(3) = 2400" sets servo 3 *
*          : to its maximum point.                                             *
*          :                                                                   *
*          : Call Sub "Servo.On" to enable Interrupt and start sending pulses  *
*          : to servos.  Call "Servo.Off" to stop Interrupt firing and stop    *
*          : sending pulses to servos.                                         *
*          :                                                                   *
*          : Flag "ServoInterruptComplete" is set to TRUE when Interrupt code  *
*          : has just finished.  Main Program may poll this flag to know when  *
*          : it is safe to run timing-sensitive code.                          *
*          :                                                                   *
*          : Uses Timer1 and Compare module to trigger interrupt at correct    *
*          : intervals for servo pulses.  Works with clock speeds of 4, 8, 16, *
*          : and 32MHz.                                                        *
*          :                                                                   *
********************************************************************************
}

Module Servo 

'===============================================================================
'User definable options
'-------------------------------------------------------------------------------

#option Servo_NumberOfServos = 8                    'Number of servos (1 - 8)
#option Servo_Priority = ipHigh                     'Servo interrupt priority

'===============================================================================
'Variable and Const Declarations
'-------------------------------------------------------------------------------

'---Set module Constants according to Option values

Private Const ServoPriority = Servo_Priority

#if Servo_NumberOfServos in (1 to 8)

    Private Const NumberOfServos As Byte = Servo_NumberOfServos
    
    #if Servo_NumberOfServos = 1                    '}
        Private Const OffMask As Byte = %11111110   '}
    #elseif Servo_NumberOfServos = 2                '}
        Private Const OffMask As Byte = %11111100   '} 
    #elseif Servo_NumberOfServos = 3                '}According to how many            
        Private Const OffMask As Byte = %11111000   '}servos are connected, set  
    #elseif Servo_NumberOfServos = 4                '}the mask which is AND'd 
        Private Const OffMask As Byte = %11110000   '}with PortB to turn off the
    #elseif Servo_NumberOfServos = 5                '}servo pins.
        Private Const OffMask As Byte = %11100000   '}    
    #elseif Servo_NumberOfServos = 6                '}
        Private Const OffMask As Byte = %11000000   '}    
    #elseif Servo_NumberOfServos = 7                '}
        Private Const OffMask As Byte = %10000000   '}     
    #else                                           '}
        Private Const OffMask As Byte = %00000000   '} 
    #endif    

#else

    #error "Servo_NumberOfServos must be in range 1 to 8"

#endif

 
#if _clock = 4                                      '}
    Private Const Timer1Config As Byte = %00000000  '}
#elseif _clock = 8                                  '}According to clock speed,
    Private Const Timer1Config As Byte = %00010000  '}set Timer1 pre-scaler to
#elseif _clock = 16                                 '}give 1 tick every 1 uS.
    Private Const Timer1Config As Byte = %00100000  '}
#elseif _clock = 32                                 '}
    Private Const Timer1Config As Byte = %00110000  '}
#else                                               
    Private Const Timer1Config As Byte = %00000000
    #warning "Servo Module only works correctly with clock of 4, 8, 16 or 32MHz"
#endif


'---Aliases for Timer1 & CCP Module Registers
Private Dim
    Timer1Value As TMR1L.AsWord,
    Timer1On As T1CON.Booleans(0),
    
    CompareValue As CCPR1L.AsWord,
    CompareInterruptFlag As PIR1.Booleans(2),
    CompareInterruptEnable As PIE1.Booleans(2),
    CompareInterruptHighPriority As IPR1.Booleans(2),
    CompareUsingTimer3 As T3CON.Booleans(3)
    
    
'---Private Variables
Private Dim
    PauseTime As Word,                      'Gap in uS between end of last servo
                                            'pulse and start of first (set so
                                            'servos are refreshed at 50Hz)                  

    ServoOnMask As Byte,                    'OR'd with PortB to turn on current
                                            'servo pin
                                            
    ServoIndex As Byte                      'Holds index number of current servo

  
'---Public Variables
Public Dim
    ServoPosition(NumberOfServos) As Word,  'Holds pulse length in uS for servos
    ServoInterruptComplete As Boolean       'Flag to show interrupt has finished

    
'===============================================================================
'Interrupts
'-------------------------------------------------------------------------------

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'Name:      OnTimer
'Purpose:   Interrupt triggered when Timer1 reaches CCPR1 register value
'           (CompareValue).
'
'Notes:     Puts servo pin high and then sets CCPR1 to desired pulse length for
'           that servo.  Next time through interrupt, next servo pin is set high
'           and CCPR1 is set to new pulse length.  Process repeats until all
'           servo pins have been pulsed.  Then, all pins off and CCPR1 set to
'           time required to wait before starting again to give 50Hz refresh.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Interrupt OnTimer(ServoPriority)
    
    Save(FSR0, FSR1)                            'Context save
    
    If ServoIndex > Bound(ServoPosition) Then   'All servo pulses have been sent
                                                    
        PORTB = PORTB And OffMask                   'Turn off all servo pins
        
        CompareValue = PauseTime                    'Time until next interrupt =
                                                    'time between cycles

        PauseTime = 20000                           '}Reset all variables for
        ServoIndex = 0                              '}next cycle
        ServoOnMask = %00000001                     '}

    Else                                        'Not all servo pulses have been
                                                'sent
    
        PORTB = PORTB And OffMask                   'Turn off all servo pins
        PORTB = PORTB Or ServoOnMask                'Turn on the next servo pin
        
        CompareValue = ServoPosition(ServoIndex)    'Time until next interrupt =
                                                    'pulse length of servo    
        
        PauseTime = PauseTime - CompareValue        'Update time between cycles
                                                    'to maintain 50Hz refresh
        
        Inc(ServoIndex)                             '}Ready for next servo
        ServoOnMask = ServoOnMask << 1              '}
        
    End If
    
    Restore                                      'Context restore

    CompareInterruptFlag = False                 'Clear interrupt flag
    
    ServoInterruptComplete = True                'Set flag to let main program
                                                 'know interrupt has finished
End Interrupt



'===============================================================================
'Public Subs and Functions
'-------------------------------------------------------------------------------

'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'Name:      On
'Purpose:   Start sending servo control pulses
'
'Notes:     Resets Timer1 and turns it on, allowing it to trigger the Interrupt
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Public Sub On()

    PORTB = PORTB And OffMask           'Turn off all servo pins
    
    PauseTime = 20000                   '}
    ServoIndex = 0                      '}Set initial values for variables
    ServoOnMask = %00000001             '}
    
    CompareValue = PauseTime            'Set time in uS before interrupt fires
    
    Timer1Value = 0                     'Reset Timer1
    Timer1On = True                     'Start Timer1

End Sub


'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'Name:      Off
'Purpose:   Stop sending servo control pulses
'
'Notes:     Turns off Timer1, preventing Interrupt from firing.
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Public Sub Off()

    Timer1On = False                    'Turn off Timer1
    PORTB = PORTB And OffMask           'Turn off all servo pins
    
End Sub


'===============================================================================
'Module Initialisation
'-------------------------------------------------------------------------------

For ServoIndex = 0 To Bound(ServoPosition)  '}Set all servos to mid-point (pulse
    ServoPosition(ServoIndex) = 1500        '}length = 1500uS)
Next                                        '}

TRISB = TRISB And OffMask                   'Set servo pins to outputs

'Initialise CCP and Timer1 modules:
CompareUsingTimer3 = False                  'Compare module uses Timer1

#if Servo_Priority = ipHigh Then            '}
    CompareInterruptHighPriority = True     '}
#else                                       '}Set interrupt priority
    CompareInterruptHighPriority = False    '}
#endif                                      '}
                                         
T1CON = Timer1Config                        'Set pre-scaler, Timer1 off

CCP1CON = %00001011                         'Set special event trigger (Resets
                                            'Timer1 and triggers interrupt when
                                            'Timer1 value reaches CCPR1 value)

CompareInterruptEnable = True               'Enable interrupt on Timer1 = CCPR1

Enable(OnTimer)                             'Enable interrupt routine





